#!/usr/bin/python3

import unittest
import os
import sys

# Add parent to the start of os.path, import ucammap
sys.path.insert(0,os.path.join(os.path.dirname(__file__),"..")) 
import ucammap

address_data = [
    {'type': 'site',     'site': 'New Museums Site'}, 
    {'type': 'number',   'number': '123'},
    {'type': 'number',   'number': '456'},
    {'type': 'street',   'street': 'Pembroke Street'}, 
    {'type': 'place',    'place': 'Cambridge'}, 
    {'type': 'postcode', 'postcode': 'CB2 3QH'},
    {'type': 'number',   'number': '999'},
    {'type': 'address',  'address': 'Some unparsed text'},
    {'type': 'location', 'location': {
        'type': 'building', 
        'ref': 'H132',
        'lat': 52.17525, 
        'lon': 0.141464, 
        'id': 1643255315 }
    },
    {'type': 'entrance', 'entrance': {
        'type': 'entrance', 
        'ref': 'H131-A',
        'name': "Addenbrooke's Hospital Main Entrance",
        'lat': 52.175206, 
        'lon': 0.140538,
        'direction': 104, 
        'user': 'f', 
        'primitive': 'n', 
        'id': 1643254795, 
        'entrance': 'main', 
        'door': True }
    },
    {'type': 'entrance', 'entrance': {
        'type': 'entrance',
        'ref': 'H-A',
        'lat': 52.176241, 
        'lon': 0.144354,
        'direction': 15, 
        'user': 'fbv', 
        'primitive': 'n', 
        'id': 664530277,
        'entrance': 'main', 
        'door': False}
    }
]

class AddressTestCase(unittest.TestCase):

    def testPopulated(self):
        """General tests on a populated address object"""

        addr = ucammap.Address(address_data)

        self.assertEqual (addr.raw_data, address_data)

        self.assertTrue  (isinstance(addr.as_block(),list))
        self.assertEqual (len(addr.as_block()),5)

        self.assertEqual (addr.as_line(),
           'New Museums Site, 123 456 Pembroke Street, Cambridge, CB2 3QH, 999')

        self.assertEqual (str(addr),
           'New Museums Site, 123 456 Pembroke Street, Cambridge, CB2 3QH, 999')

        self.assertTrue(isinstance(addr.location(),ucammap.Entity))

        self.assertTrue(isinstance(addr.entrances(),list))
        self.assertEqual(len(addr.entrances()),2)

    def testEmpty(self):
        """General tests on an empty address"""

        empty = ucammap.Address([])

        self.assertEqual (empty.raw_data, [])

        self.assertTrue  (isinstance(empty.as_block(),list))
        self.assertEqual (len(empty.as_block()),0)

        self.assertEqual (empty.as_line(),'')

        self.assertEqual (empty.location(),None)

        self.assertTrue(isinstance(empty.entrances(),list))
        self.assertEqual(empty.entrances(),[])

    def testUnstructured(self):
        """Tests on an address only containing unstructured data"""

        partial_data = [
            {'type': 'location', 'location': 987},
            {'type': 'address',  'address': 'Some unparsed text'}
            ]
        partial = ucammap.Address(partial_data)

        self.assertEqual (partial.raw_data, partial_data)
        self.assertTrue  (isinstance(partial.as_block(),list))
        self.assertEqual (len(partial.as_block()),0)
        self.assertEqual (partial.as_line(),'Some unparsed text')

    def testExceptions(self):
        """Test address exceptions"""

        # A address containing an entry with no type throws an error
        
        t = lambda: ucammap.Address([{'typeZZZ': 'location', 'location': 987}])
        self.assertRaises(ucammap.APIError,t)

        # A address containing an entry with no data field throws an error
        t = lambda: ucammap.Address([{'type': 'location', 'locationZZZ': 987}])
        self.assertRaises(ucammap.APIError,t)


if __name__ == '__main__':
    unittest.main()
