#!/usr/bin/python3

import unittest
import os
import sys

# Add parent to the start of os.path, import ucammap
sys.path.insert(0,os.path.join(os.path.dirname(__file__),"..")) 
import ucammap

# Note that this is an 'impossible' entity that exhibits all the
# attribures of all the possible entities

entity_data = {
    'direction': 123.4,
    'door': True,
    'id': 1091639792,
    'inst': 'cs', 
    'lat': 52.203379, 
    'lon': 0.119701, 
    'name': 'Computing Service', 
    'options': 'foo',
    'prefix': 'University', 
    'primitive': 'n', 
    'ref': 'M-A', 
    'role': 'nonacademic', 
    'sd': 'ucs',
    'sortasvia': 42,
    'suffix': 'UCS',
    'type': 'institution', 
    'user': 'fbvg',
    'institution': {
        'type': 'institution', 
        'inst': 'misd', 
        'name': 'MISD', 
        },
    'parent': {
        'type': 'institution', 
        'inst': 'misd', 
        'name': 'MISD', 
    },
    'entrances': [
        {'entrance': 'main', 
         'type': 'entrance', 
        }, 
    ],
    'occupants': [
        {'type': 'institution', 
         'inst': 'misd', 
         'name': 'MISD', 
         },
    ],
    'parents': [
        {'type': 'institution', 
         'inst': 'misd', 
         'name': 'MISD', 
         },
    ],
    'info': [
        {'type': 'phone',   'phone': '+44 1223 334600'}, 
    ],
}

simple_attributes = set(['direction', 'door', 'id', 'inst', 'lat', 'lon',
                     'name', 'options', 'prefix', 'primitive', 'ref', 
                     'role', 'sd', 'sortasvia', 'suffix', 'type', 'user'])

entity_attributes = (['institution', 'parent'])

entity_list_attributes = (['entrances', 'occupants', 'parents'])

all_attributes = simple_attributes.union(entity_attributes, 
                                         entity_list_attributes,
                                         set(['info']))

class EntityTestCase(unittest.TestCase):

    def setUp(self):
        self.entity = ucammap.Entity(entity_data)
        self.empty  = ucammap.Entity({})

    def testType(self):
        self.assertTrue(isinstance(self.entity,ucammap.Entity))

    def testRawData(self):
        """Extracting entity raw data"""
        self.assertEqual(self.entity.raw_data,entity_data)
        self.assertEqual(self.empty.raw_data,{})

    def testSimpleAttributes(self):
        """Retrieving simple entity attributes"""

        for key in simple_attributes:
          self.assertEqual(getattr(self.entity,key),entity_data[key],msg=key)

        t = lambda x: self.entity.nonsuch
        self.assertRaises(AttributeError, t, None) 

    def testEntityAttributes(self):
        """Attributes that return more Entities"""

        for ent in entity_attributes:
            self.assertTrue(
                isinstance(getattr(self.entity,ent),ucammap.Entity), msg=ent)

    def testEntityListAttributes(self):
        """Attributes that return lists of Entities"""

        for ent in entity_list_attributes:
            self.assertTrue(
                isinstance(getattr(self.entity,ent),list), msg=ent)
            self.assertTrue(
                isinstance(getattr(self.entity,ent)[0],ucammap.Entity),msg=ent)

    def testInfo(self):
            
        self.assertTrue(isinstance(self.entity.info,list))
        self.assertTrue(isinstance(self.entity.info[0],ucammap.Info))

    def testEmpty(self):
        """(Not) retrieving attributed from 'empty'"""

        for ent in simple_attributes:
            self.assertEqual(getattr(self.empty,ent),None,ent)

        for ent in entity_attributes:
            self.assertEqual(getattr(self.empty,ent),None,ent)
            
        for ent in entity_list_attributes:
            self.assertEqual(getattr(self.empty,ent),[],ent)

        self.assertEqual(self.empty.info,[])

        t = lambda x: self.entity.nonsuch
        self.assertRaises(AttributeError,t, None) 

    def testMethods(self):
        """Entity utility methods"""

        # full_name
        self.assertEqual (self.entity.full_name(),
                           'University Computing Service (UCS)')
        self.assertEqual (self.empty.full_name(),'')

        # sort_name
        self.assertEqual (self.entity.sort_name(),
                           'Computing Service, University (UCS)')
        self.assertEqual (self.empty.sort_name(),'')

        # attributes
        self.assertEqual(set(self.entity.attributes()),all_attributes)

        # filtered_info
        info = self.entity.filtered_info('phone')
        self.assertTrue (isinstance(info,list))
        self.assertEqual(len(info),1)
        self.assertTrue (isinstance(info[0],ucammap.Info))

        info = self.empty.filtered_info('phone')
        self.assertTrue (isinstance(info,list))
        self.assertEqual(len(info),0)

        # info_types
        types = self.entity.info_types()
        self.assertTrue (isinstance(types,list))
        self.assertEqual(len(types),1)
        self.assertEqual(types[0],'phone')

        types = self.empty.info_types()
        self.assertTrue (isinstance(types,list))
        self.assertEqual(len(types),0)

        # __str__
        self.assertEqual (str(self.entity), 
                          'University Computing Service (UCS)')

if __name__ == '__main__':
    unittest.main()
