"""
A Python interface to the University of Cambridge Map database API.

This is a Python interface to the database API to the University of
Cambridge Map. It is expected to work in Python 2.6 and above, and
Pythion 3. The underlying API is described at

  https://wiki.cam.ac.uk/university-map/The_Database_API

The code contains extensive in-line documentation so you may want to
look at that - in the source, via the built in Python help() function,
or in the ucammap.pydoc file included in the distribution.

The interface is implemented by the 'ucammap' module which should be
imported in the usual way. The primary (close to only) way to use this
is to create a ucammap.API object and then to call its search()
method. All keyword arguments to search() are converted to URL
parameters and included in a call to the underlying interface - see
the API documentation for more detail of what's possible. The result
is a (possibly empty) sequence of objects representing the selected
entities. Information about institutions, sites, buildings, etc. are
represented by ucammap.Entity objects, and contact information records
and addresses by ucammap.Info and ucammap.Address objects.

ucammap.API objects also provide an inflate() method to retrieve a
full copy of an Entity object based on a partial one.  The API class
uses the standard logging module (at 'info' and 'debug') to track the
retrieval of API data.

Access to information stored in Entity objects is available via
attributes named after the coresponding keys described in the
API. These return either simple values (string, integer, float,
boolean), single instances of other ucammap objects, or
possibably-empty sequences of other objects. Entity objects provide
various convinience utility methods, such as
ucammap.Entity.sort_name() which returns the entity's name in a form
suitable for use in sorted lists.

ucammap.Info objects, representing a single row of contact
information, make information available via the attributes 'type' and
'value', sometimes suplimented by 'role' and 'attributes'. Values are
either strings or ucammap.Address objects.

ucammap.Address objects provide methods for extracting address
information in various formats.

Supported attributes that have no value in a particular object return
None for simple values or empty sequences.

Note the following:
  * All objects produce something useful when converted to a string
    (and so when printed), which is useful for experimentation
  * All data attributes evaluate as False if not available, making 
    this pattern easy -
       if object.thing:
           print(<header for thing>)
           <do stuff with thing>
  * The raw data that initialized any particular object is available
    as its raw_data attribute

Basic usage:
>>> import ucammap
>>> api = ucammap.API()
>>> for result in api.search(q = 'computing'):
...     print(result.full_name())
... 
Department of Applied Mathematics and Theoretical Physics (DAMTP)
Bioinformatics and Computational Biology Services
Cambridge Computational Biology Institute (CCBI)
Department of Chemistry
Collaborative Computing Project for NMR
Computer Laboratory
University Computing Service (UCS)
High Performance Computing Service (HPCS)

>>> import ucammap
>>> api = ucammap.API()
>>> for entity in api.search(inst = 'cs'):
...   for info in entity.filtered_info('address'):
...     print(info.value.as_line())
... 
West Cambridge Site, Roger Needham Building, 7 J J Thomson Avenue, Cambridge, CB3 0RB
New Museums Site, Cockcroft Building, Pembroke Street, Cambridge, CB2 3QH

This module implements the following classes:

  API       -- an API for querying map database
  Entity    -- a representation of the main entities (institutions, 
               buildings, etc. returned by the api
  Info      -- a representation of contact information
  Address   -- a representation of address information
  APIError  -- a local exception object for the class 

"""

# Patch up urllib imports for 2.6 compatibility
try:
    from urllib.request import urlopen
    from urllib.error import HTTPError
    from urllib.parse import urlencode
except ImportError:
    from urllib2 import urlopen
    from urllib2 import HTTPError
    from urllib import urlencode

import json

import logging
logger = logging.getLogger(__name__)


class API:

    """
    An interface to the Map API.

    Provides basic searching and retrieval functions for accessing
    information from the database API.

    Implements:
       __init__ 
       search   -- Search for entities
       inflate  -- Return a full Entity object based on a partial 
                   one returned previously

    """

    def __init__(self, base='http://map.cam.ac.uk/v6.json'):
        """
        Initialise the interface.

        Keyword arguments:
        base -- Base URL of the interface 
                (default 'http://map.cam.ac.uk/v5.json')

        """
        self.base = base

    def search(self, **kwargs):
        """
        Search database, return matching entities.

        Keyword arguments: all keyword arguments are converted to URL
        parameters and included in a call to the underlying interface.

        Returns a (possibly empty) sequence of Entity objects.

        Raises ucammap.APIError following error return from the
        interface. May propagates exceptions from urllib if the
        underlying HTTP request fails.

        """
        
        parameters  = urlencode(kwargs)
        logger.info("search called with %s",kwargs) 

        response = None
        try:
            url = self.base + '?' + parameters
            logger.debug("Requesting %s",url)
            response = urlopen(url)
        except HTTPError as e:
            if e.code == 400 and e.info()["Content-type"] == 'application/json':
                # Status = 400 + JSON content is API error - use content
                response = e
            else:
                # ...otherwise propagate
                raise

        logger.debug("Response code %s, type %s", 
                    response.code,response.info()["Content-type"])   

        # Something badly wrong if response not application/json  
        content_type = response.info()["Content-type"]
        if content_type != 'application/json':
            raise APIError('Unexpected content returned: ' + content_type)

        # Encoding not specified - assume default (utf-8)
        data = json.loads(response.read().decode('utf-8'))
        logger.debug("Response content %s",data)

        results = []
        for item in data:
            if item['type'] == 'error':
                # Convert returned error entities to exceptions
                raise APIError(item['error'])
            else:
                # ...otherwise convert to entity
                results.append(Entity(item))

        logger.info("Successfully returned %s entities",len(results))        

        return results

    def inflate(self,entity):
        """
        Return a full Entity based on a partial one returned previously.

        Given an entity, extract its inst or ref key and retrieve all
        the available information about it. This is intended
        particularly for expanding the partial 2nd level entities that
        are returned as part of other entities.

        Arguements:
          entity -- the entity to expand        

        Returns the expanded entity. Note that the returned value is
        normally a new Entity object - the method doesn't inflate the
        existing one in place. However if an expanded entity can't be found,
        the method returns the entity it was called with.

        Raises TypeError if called with an object that supports
        neither 'inst' nor 'ref' attributes. Raises ucammap.APIError
        if the resulting search returns more than one entity.

        """

        if hasattr(entity,'inst') and entity.inst:
            result = self.search(inst = entity.inst)
        elif hasattr(entity,'ref') and entity.ref:
            result = self.search(ref = entity.ref)
        else:
            raise TypeError("inflate - wrong parameter type")

        if len(result) > 1:
            raise APIError("inflate - more than one result")
        
        if result:
            return result[0]
        else:
            return entity


class Entity:

    """
    A representation of a database entity (institution, building, etc.).

    Implements:
      __init__()
      attributes()
      full_name()
      sort_name()
      fintered_info()
      info_types()
      __str__()
    
    Instance variables:
      raw_data -- underlying data structure returned by the API 
                  as a dictionary
      direction
      door
      id
      inst
      lat  
      lon
      name
      options
      prefix 
      primitive
      ref
      role
      sd
      sortasvia
      suffix
      type
      user
      institution
      parent
      entrances
      occupants
      parents
      info
    
     """

    # Set literal replaced with set() for 2.6 compatibility
    valid_attributes = set(['direction', 'door', 'id', 'inst', 'lat', 'lon',
                        'name', 'options', 'prefix', 'primitive', 'ref', 
                        'role', 'sd', 'sortasvia', 'suffix', 'type', 'user',
                        'institution', 'parent', 'entrances', 'occupants', 
                        'parents','info'])

    def __init__(self,data):
        """Initialise new entity object from dictionary representation"""
        self.raw_data = data

    def __getattr__(self,name):
 
        """
        Return any API data element as an attribute.

        Arguements:
          name -- the name of the element required

        Returns the data requested. Simple elements are returned as
        strings, or None if not present; elements relating to single
        entities ('parent', 'institution') are returned as an Entity
        object, or None of not present; elements relating to multiple
        entities ('entrances', 'occupants', 'parents') are returned as
        a (possibly empty) sequence of Entity objects; references to
        contact info ('info') are returned as a (possibly empty)
        sequence of Info objects.

        Note that all elements evaluate as false if not available,
        making
 
          if entity.thing:
            <do stuff with thing> 

        easy.

        """

        if name not in self.valid_attributes:
            raise AttributeError("'{0}' object has no attribute '{1}'"
                                 .format(self.__class__,name))

        # All the fields that are sequences of entities
        if name in ('entrances', 'occupants', 'parents'):
            entities = []
            if name in self.raw_data:
                for item in self.raw_data[name]:
                    entities.append(Entity(item))
            return entities

        # All the fields that are single entities
        elif name in ('parent','institution') and name in self.raw_data: 
            return (Entity(self.raw_data[name]))

        # Info, which is a sequence of info objects
        elif name == 'info':
            info = []
            if 'info' in self.raw_data:
                for item in self.raw_data['info']:
                    info.append(Info(item))
            return info

        # Simple fields
        elif name in self.raw_data:
            return self.raw_data[name]

        # Otherwise None
        else:
            return None

    def attributes(self):
        """Return a list of the attributes actually in the entity."""
        return list(self.raw_data.keys())

    def full_name(self):
        """Return a full human-readable name."""
        value = []
        if self.prefix:
            value.append(self.prefix)
        if self.name:
            value.append(self.name)
        if self.suffix:
            value.append('(' + self.suffix + ')')
        return ' '.join(value)   

    def sort_name(self):
        """Return a human-readable name suitable for sorted lists."""
        value = []
        if self.name:
            value.append(self.name)
        if self.prefix:
            value.append(self.prefix)
        value = [', '.join(value)]
        if self.suffix:
            value.append('(' + self.suffix + ')')
        return ' '.join(value)    

    def filtered_info(self,key):
        """Return all Info records of type matching the supplied key."""
        return [c for c in self.info if c.type == key]

    def info_types(self):
        """Return a sorted list of unique Info record types for the entity."""
        s = set([info.type for info in self.info])
        return sorted(list(s))

    def __str__(self):
        """Return a human-readable representation."""
        return self.full_name()


class Info:
    """A container for a single contact info item.

    Implements:
      __init__
      __str__
  
    Instance variables:
      raw_data -- underlying data structure returned by the API as a 
                  sequence of dicts
      type     -- the type of info recod (see API documentation) 
      value    -- the value coresponding to 'type'
      role     -- coresponding 'role'
      options  -- coresponding 'options'
                
      Where 'type' is 'address' the returned value will be a
      ucammap.Address object, all other retuned values are strings.

    """

    # Set literal replaced with set() for 2.6 compatibility
    valid_attributes = set(['type','value','role','options'])

    def __init__(self,data):
        """
        Initialise a new Info object from dictionary representation.

        """

        self.raw_data = data

    def __getattr__(self,name):
        """Return attributes from an Info record.

        Raises APIError if called with name='value' on an info 
        record that doesn't contain a 'type' field or which doesn't 
        contain a data field named after the type field.
        """

        if name not in self.valid_attributes:
            raise AttributeError("'{0}' object has no attribute '{1}'"
                                 .format(self.__class__,name))

        if name == 'value':
            if not 'type' in self.raw_data:
                raise APIError('Type field missing from info record ' + str(self.raw_data))
            info_type = self.raw_data['type']
            if not info_type in self.raw_data:
                raise APIError('Data field "' + info_type + 
                               '" missing from info record ' + str(self.raw_data))
            if info_type == 'address':
                return Address(self.raw_data['address'])
            else:
                return self.raw_data[info_type]
        elif name in self.raw_data:
            return self.raw_data[name]
        else:
            return None
        
    def __str__(self):
        """Return a useful string representation of value + role"""
        result = self.type.capitalize() + ': ' + str(self.value)
        if self.role:
            result += ' (' + self.role + ')'
        return result    


class Address:

    """
    A representation of a structured address.

    Implements:
      __init__ 
      lines()     -- lines of the address as a sequence
      location()  -- location object associated with address, or None
      entranes()  -- sequence of entrances associated with location, or []
      as_line()   -- address as one line
      as_block()  -- address as a sequence of lines
      __str__

    Instance variables:
      raw_data  -- the underlying API data structure (a list of two 
                   element dicts)
 
    """

    def __init__(self,address):
        """
        Initialise address object from a sequence of dictionaries

        Raises APIError if called with a list element that doesn't
        contain a 'type' field or which doesn't contain a data field
        named after the type field.

        """

        for line in address:
            if not 'type' in line:
                raise APIError('Type field missing from address record ' 
                               + str(line))
            address_type = line['type']
            if not address_type in line:
                raise APIError('Data field "' + address_type + 
                               '" missing from address record ' + str(line))
        self.raw_data = address

    def lines(self):
        """Return address lines as a sequence of (type,value) pairs."""
        result  = []
        for line in self.raw_data:
            address_type = line['type']
            result.append((address_type,line[address_type]))
        return result

    def location(self):
        """Return the location object (building/site/college entity) 
           associated with this address, or None if there isn't one."""
        for (key,value) in self.lines():
            # return first - spec says 'only one'
            if key == 'location':
                return Entity(value)
        return None

    def entrances(self):
        """Return a (possibably empty) sequence of entrance objects
           associated with the location object of this address."""
        entities = []
        for (key,value) in self.lines():
            if key == 'entrance':
                entities.append(Entity(value))
        return entities

    def as_block(self):
        """Returns address as an list of strings, one per line."""
        result  = []
        numbers = []
        for (key,value) in self.lines():
            if key in ('location', 'entrance', 'address'):
                pass
            elif key == 'number':
                numbers.append(value)
            else:
                if numbers:
                    result.append(' '.join(numbers) + ' ' + value)
                    numbers = []
                else:
                    result.append(value)
        if numbers:
            result.append(' '.join(numbers))
        return result

    def as_line(self):
        """Returns address as a single string. Could be empty."""
        block = self.as_block()
        if block:
            return ", ".join(block)
        # ... otherwise return 'address' if available
        for (key,value) in (self.lines()):
            if key == 'address':
                return value
        # ... otherwise an empty string
        return ''

    def __str__(self):
        """Return a string represenattion of an address."""
        return self.as_line()


class APIError(Exception):
    """Custom error class for the ucammap module."""

    def __init__(self,reason):
        self.reason = reason

    def __str__(self):
        return self.reason


if __name__ == "__main__":
    import doctest
    doctest.testmod()
